@@ -1,16 +1,21 @@
-/* Copyright 2002-2007 Justin Erenkrantz and Greg Stein
+/* ====================================================================
+ *    Licensed to the Apache Software Foundation (ASF) under one
+ *    or more contributor license agreements.  See the NOTICE file
+ *    distributed with this work for additional information
+ *    regarding copyright ownership.  The ASF licenses this file
+ *    to you under the Apache License, Version 2.0 (the
+ *    "License"); you may not use this file except in compliance
+ *    with the License.  You may obtain a copy of the License at
  *
- * Licensed under the Apache License, Version 2.0 (the "License");
- * you may not use this file except in compliance with the License.
- * You may obtain a copy of the License at
+ *      http://www.apache.org/licenses/LICENSE-2.0
  *
- *     http://www.apache.org/licenses/LICENSE-2.0
- *
- * Unless required by applicable law or agreed to in writing, software
- * distributed under the License is distributed on an "AS IS" BASIS,
- * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
- * See the License for the specific language governing permissions and
- * limitations under the License.
+ *    Unless required by applicable law or agreed to in writing,
+ *    software distributed under the License is distributed on an
+ *    "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
+ *    KIND, either express or implied.  See the License for the
+ *    specific language governing permissions and limitations
+ *    under the License.
+ * ====================================================================
  */
 
 #include <stdlib.h>
@@ -20,6 +25,8 @@
 #include <apr_strings.h>
 #include <apr_version.h>
 
+#include <openssl/opensslv.h>
+
 #include "serf.h"
 #include "serf_private.h"
 
@@ -392,8 +399,8 @@ static void test_keepalive_limit_one_by_one(CuTest *tc)
     CuAssertIntEquals(tc, APR_SUCCESS, status);
 
     for (i = 0 ; i < SEND_REQUESTS ; i++) {
-        create_new_request_with_resp_hdlr(tb, &handler_ctx[i], "GET", "/", i+1,
-                                          handle_response_keepalive_limit);
+        create_new_request_ex(tb, &handler_ctx[i], "GET", "/", i+1,
+                              NULL, handle_response_keepalive_limit);
         /* TODO: don't think this needs to be done in the loop. */
         serf_connection_set_max_outstanding_requests(tb->connection, 1);
     }
@@ -523,8 +530,8 @@ static void test_keepalive_limit_one_by_one_and_burst(CuTest *tc)
     CuAssertIntEquals(tc, APR_SUCCESS, status);
 
     for (i = 0 ; i < SEND_REQUESTS ; i++) {
-        create_new_request_with_resp_hdlr(tb, &handler_ctx[i], "GET", "/", i+1,
-                                          handle_response_keepalive_limit_burst);
+        create_new_request_ex(tb, &handler_ctx[i], "GET", "/", i+1,
+                              NULL, handle_response_keepalive_limit_burst);
         serf_connection_set_max_outstanding_requests(tb->connection, 1);
     }
 
@@ -646,7 +653,6 @@ static void test_connection_userinfo_in_url(CuTest *tc)
     handler_baton_t handler_ctx[2];
     const int num_requests = sizeof(handler_ctx)/sizeof(handler_ctx[0]);
     int i;
-    progress_baton_t *pb;
 
     test_server_message_t message_list[] = {
         {CHUNKED_REQUEST(1, "1")},
@@ -1016,6 +1022,22 @@ static const char *all_server_certs[] = {
     "test/server/serfrootcacert.pem",
     NULL };
 
+static const char **server_certs_srcdir(const char **certs,
+                                         apr_pool_t *result_pool)
+{
+  const char **result;
+  int i = 0;
+  while (certs[i])
+    i++;
+
+  result = apr_pcalloc(result_pool, sizeof(result[0]) * (i + 1));
+
+  while (i-- > 0)
+    result[i] = get_srcdir_file(result_pool, certs[i]);
+
+  return result;
+}
+
 static apr_status_t validate_servercert(const serf_ssl_certificate_t *cert,
                                         apr_pool_t *pool)
 {
@@ -1106,34 +1128,90 @@ static apr_status_t validate_rootcacert(const serf_ssl_certificate_t *cert,
     return APR_SUCCESS;
 }
 
-static apr_status_t
-ssl_server_cert_cb_expect_failures(void *baton, int failures,
-                                   const serf_ssl_certificate_t *cert)
+static const char *format_cert_failures(int failures, apr_pool_t *pool)
 {
-    test_baton_t *tb = baton;
-    int expected_failures = *(int *)tb->user_baton;
+    const char *str = "";
 
-    tb->result_flags |= TEST_RESULT_SERVERCERTCB_CALLED;
+    if (failures & SERF_SSL_CERT_NOTYETVALID) {
+        str = apr_pstrcat(pool, str, *str ? "|" : "", "CERT_NOTYETVALID", NULL);
+        failures &= ~SERF_SSL_CERT_NOTYETVALID;
+    }
 
-    /* We expect an error from the certificate validation function. */
-    if (failures & expected_failures)
-        return APR_SUCCESS;
+    if (failures & SERF_SSL_CERT_EXPIRED) {
+        str = apr_pstrcat(pool, str, *str ? "|" : "", "CERT_EXPIRED", NULL);
+        failures &= ~SERF_SSL_CERT_EXPIRED;
+    }
+
+    if (failures & SERF_SSL_CERT_UNKNOWNCA) {
+        str = apr_pstrcat(pool, str, *str ? "|" : "", "CERT_UNKNOWNCA", NULL);
+        failures &= ~SERF_SSL_CERT_UNKNOWNCA;
+    }
+
+    if (failures & SERF_SSL_CERT_SELF_SIGNED) {
+        str = apr_pstrcat(pool, str, *str ? "|" : "", "CERT_SELF_SIGNED", NULL);
+        failures &= ~SERF_SSL_CERT_SELF_SIGNED;
+    }
+
+    if (failures & SERF_SSL_CERT_UNKNOWN_FAILURE) {
+        str = apr_pstrcat(pool, str, *str ? "|" : "", "CERT_UNKNOWN_FAILURE", NULL);
+        failures &= ~SERF_SSL_CERT_UNKNOWN_FAILURE;
+    }
+
+    if (failures & SERF_SSL_CERT_REVOKED) {
+        str = apr_pstrcat(pool, str, *str ? "|" : "", "CERT_REVOKED", NULL);
+        failures &= ~SERF_SSL_CERT_REVOKED;
+    }
+
+    if (failures) {
+        serf__log(TEST_VERBOSE, __FILE__, "Unexpected or unknown cert failure\n");
+        abort();
+    }
+
+    if (*str)
+        return str;
     else
-        return SERF_ERROR_ISSUE_IN_TESTSUITE;
+        return "NONE";
 }
 
-static apr_status_t
-ssl_server_cert_cb_expect_allok(void *baton, int failures,
-                                const serf_ssl_certificate_t *cert)
+/* Logs failures in tb->user_baton, for later validation. */
+static apr_status_t ssl_server_cert_cb_log(void *baton, int failures,
+                                           const serf_ssl_certificate_t *cert)
 {
     test_baton_t *tb = baton;
+    const char *cert_str;
+
     tb->result_flags |= TEST_RESULT_SERVERCERTCB_CALLED;
 
-    /* No error expected, certificate is valid. */
-    if (failures)
-        return SERF_ERROR_ISSUE_IN_TESTSUITE;
-    else
-        return APR_SUCCESS;
+    if (cert) {
+        apr_hash_t *subject;
+        const char *common_name;
+        int depth;
+
+        subject = serf_ssl_cert_subject(cert, tb->pool);
+        if (!subject)
+            return SERF_ERROR_ISSUE_IN_TESTSUITE;
+
+        common_name = apr_hash_get(subject, "CN", APR_HASH_KEY_STRING);
+        depth = serf_ssl_cert_depth(cert);
+
+        cert_str = apr_psprintf(tb->pool, "(CN=%s, depth=%d)", common_name, depth);
+    } else {
+        cert_str = "(null)";
+    }
+
+    if (!tb->user_baton)
+        tb->user_baton = "";
+
+    tb->user_baton = apr_pstrcat(
+        tb->pool,
+        tb->user_baton,
+        "cert_cb: "
+        "failures = ", format_cert_failures(failures, tb->pool),
+        ", cert = ", cert_str,
+        "\n",
+        NULL);
+
+    return APR_SUCCESS;
 }
 
 static apr_status_t
@@ -1150,7 +1228,6 @@ static void test_ssl_handshake(CuTest *tc)
     test_baton_t *tb;
     handler_baton_t handler_ctx[1];
     const int num_requests = sizeof(handler_ctx)/sizeof(handler_ctx[0]);
-    int expected_failures;
     apr_status_t status;
     test_server_message_t message_list[] = {
         {CHUNKED_REQUEST(1, "1")},
@@ -1170,24 +1247,37 @@ static void test_ssl_handshake(CuTest *tc)
                                      message_list, num_requests,
                                      action_list, num_requests, 0,
                                      NULL, /* default conn setup */
-                                     "test/server/serfserverkey.pem",
-                                     server_cert,
+                                     get_srcdir_file(test_pool, "test/server/serfserverkey.pem"),
+                                     server_certs_srcdir(server_cert, test_pool),
                                      NULL, /* no client cert */
-                                     ssl_server_cert_cb_expect_failures,
+                                     ssl_server_cert_cb_log,
                                      test_pool);
     CuAssertIntEquals(tc, APR_SUCCESS, status);
 
-    /* This unknown failures is X509_V_ERR_UNABLE_TO_VERIFY_LEAF_SIGNATURE, 
-       meaning the chain has only the server cert. A good candidate for its
-       own failure code. */
-    expected_failures = SERF_SSL_CERT_UNKNOWNCA |
-                        SERF_SSL_CERT_UNKNOWN_FAILURE;
-    tb->user_baton = &expected_failures;
-
     create_new_request(tb, &handler_ctx[0], "GET", "/", 1);
 
     test_helper_run_requests_expect_ok(tc, tb, num_requests, handler_ctx,
                                        test_pool);
+
+    /* OpenSSL 1.1.1i allows to continue verification for certificates with an
+       unknown CA. See https://github.com/openssl/openssl/issues/11297.
+
+       These unknown failures are X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY
+       and X509_V_ERR_UNABLE_TO_VERIFY_LEAF_SIGNATURE. The second one means that
+       the chain has only the server cert. A good candidate for its own failure
+       code. */
+#if OPENSSL_VERSION_NUMBER >= 0x1010109fL /* >= 1.1.1i */
+    CuAssertStrEquals(tc,
+        "cert_cb: failures = CERT_UNKNOWNCA, cert = (CN=localhost, depth=0)\n"
+        "cert_cb: failures = CERT_UNKNOWNCA, cert = (CN=localhost, depth=0)\n"
+        "cert_cb: failures = NONE, cert = (CN=localhost, depth=0)\n",
+        tb->user_baton);
+#else
+    CuAssertStrEquals(tc,
+        "cert_cb: failures = CERT_UNKNOWNCA, cert = (CN=localhost, depth=0)\n"
+        "cert_cb: failures = CERT_UNKNOWNCA, cert = (CN=localhost, depth=0)\n",
+        tb->user_baton);
+#endif
 }
 
 /* Set up the ssl context with the CA and root CA certificates needed for
@@ -1209,7 +1299,8 @@ https_set_root_ca_conn_setup(apr_socket_t *skt,
         return status;
 
     status = serf_ssl_load_cert_file(&rootcacert,
-                                     "test/server/serfrootcacert.pem",
+                                     get_srcdir_file(pool,
+                                               "test/server/serfrootcacert.pem"),
                                      pool);
     if (status)
         return status;
@@ -1242,10 +1333,10 @@ static void test_ssl_trust_rootca(CuTest *tc)
                                      message_list, num_requests,
                                      action_list, num_requests, 0,
                                      https_set_root_ca_conn_setup,
-                                     "test/server/serfserverkey.pem",
-                                     server_certs,
+                                     get_srcdir_file(test_pool, "test/server/serfserverkey.pem"),
+                                     server_certs_srcdir(server_certs, test_pool),
                                      NULL, /* no client cert */
-                                     ssl_server_cert_cb_expect_allok,
+                                     ssl_server_cert_cb_log,
                                      test_pool);
     CuAssertIntEquals(tc, APR_SUCCESS, status);
 
@@ -1253,6 +1344,9 @@ static void test_ssl_trust_rootca(CuTest *tc)
 
     test_helper_run_requests_expect_ok(tc, tb, num_requests, handler_ctx,
                                        test_pool);
+    CuAssertStrEquals(tc,
+        "cert_cb: failures = NONE, cert = (CN=localhost, depth=0)\n",
+        tb->user_baton);
 }
 
 /* Validate that when the application rejects the cert, the context loop
@@ -1279,8 +1373,8 @@ static void test_ssl_application_rejects_cert(CuTest *tc)
                                      message_list, num_requests,
                                      action_list, num_requests, 0,
                                      https_set_root_ca_conn_setup,
-                                     "test/server/serfserverkey.pem",
-                                     server_certs,
+                                     get_srcdir_file(test_pool, "test/server/serfserverkey.pem"),
+                                     server_certs_srcdir(server_certs, test_pool),
                                      NULL, /* no client cert */
                                      ssl_server_cert_cb_reject,
                                      test_pool);
@@ -1345,7 +1439,7 @@ chain_rootca_callback_conn_setup(apr_socket_t *skt,
         return status;
 
     serf_ssl_server_cert_chain_callback_set(tb->ssl_context,
-                                            ssl_server_cert_cb_expect_allok,
+                                            ssl_server_cert_cb_log,
                                             cert_chain_cb,
                                             tb);
 
@@ -1376,10 +1470,10 @@ static void test_ssl_certificate_chain_with_anchor(CuTest *tc)
                                      message_list, num_requests,
                                      action_list, num_requests, 0,
                                      chain_rootca_callback_conn_setup,
-                                     "test/server/serfserverkey.pem",
-                                     server_certs,
+                                     get_srcdir_file(test_pool, "test/server/serfserverkey.pem"),
+                                     server_certs_srcdir(server_certs, test_pool),
                                      NULL, /* no client cert */
-                                     ssl_server_cert_cb_expect_allok,
+                                     ssl_server_cert_cb_log,
                                      test_pool);
     CuAssertIntEquals(tc, APR_SUCCESS, status);
 
@@ -1387,8 +1481,9 @@ static void test_ssl_certificate_chain_with_anchor(CuTest *tc)
 
     test_helper_run_requests_expect_ok(tc, tb, num_requests,
                                        handler_ctx, test_pool);
-
-    CuAssertTrue(tc, tb->result_flags & TEST_RESULT_SERVERCERTCB_CALLED);
+    CuAssertStrEquals(tc,
+        "cert_cb: failures = NONE, cert = (CN=localhost, depth=0)\n",
+        tb->user_baton);
     CuAssertTrue(tc, tb->result_flags & TEST_RESULT_SERVERCERTCHAINCB_CALLED);
 }
 
@@ -1421,7 +1516,7 @@ chain_callback_conn_setup(apr_socket_t *skt,
         return status;
 
     serf_ssl_server_cert_chain_callback_set(tb->ssl_context,
-                                            ssl_server_cert_cb_expect_allok,
+                                            ssl_server_cert_cb_log,
                                             cert_chain_all_certs_cb,
                                             tb);
 
@@ -1451,10 +1546,10 @@ static void test_ssl_certificate_chain_all_from_server(CuTest *tc)
                                      message_list, num_requests,
                                      action_list, num_requests, 0,
                                      chain_callback_conn_setup,
-                                     "test/server/serfserverkey.pem",
-                                     all_server_certs,
+                                     get_srcdir_file(test_pool, "test/server/serfserverkey.pem"),
+                                     server_certs_srcdir(all_server_certs, test_pool),
                                      NULL, /* no client cert */
-                                     ssl_server_cert_cb_expect_allok,
+                                     ssl_server_cert_cb_log,
                                      test_pool);
     CuAssertIntEquals(tc, APR_SUCCESS, status);
 
@@ -1463,7 +1558,10 @@ static void test_ssl_certificate_chain_all_from_server(CuTest *tc)
     test_helper_run_requests_expect_ok(tc, tb, num_requests,
                                        handler_ctx, test_pool);
 
-    CuAssertTrue(tc, tb->result_flags & TEST_RESULT_SERVERCERTCB_CALLED);
+    CuAssertStrEquals(tc,
+        "cert_cb: failures = CERT_SELF_SIGNED, cert = (CN=Serf Root CA, depth=2)\n"
+        "cert_cb: failures = NONE, cert = (CN=localhost, depth=0)\n",
+        tb->user_baton);
     CuAssertTrue(tc, tb->result_flags & TEST_RESULT_SERVERCERTCHAINCB_CALLED);
 }
 
@@ -1489,8 +1587,8 @@ static void test_ssl_no_servercert_callback_allok(CuTest *tc)
                                      message_list, num_requests,
                                      action_list, num_requests, 0,
                                      https_set_root_ca_conn_setup,
-                                     "test/server/serfserverkey.pem",
-                                     server_certs,
+                                     get_srcdir_file(test_pool, "test/server/serfserverkey.pem"),
+                                     server_certs_srcdir(server_certs, test_pool),
                                      NULL, /* no client cert */
                                      NULL, /* No server cert callback */
                                      test_pool);
@@ -1524,8 +1622,8 @@ static void test_ssl_no_servercert_callback_fail(CuTest *tc)
                                      message_list, num_requests,
                                      action_list, num_requests, 0,
                                      NULL, /* default conn setup, no certs */
-                                     "test/server/serfserverkey.pem",
-                                     server_certs,
+                                     get_srcdir_file(test_pool, "test/server/serfserverkey.pem"),
+                                     server_certs_srcdir(server_certs, test_pool),
                                      NULL, /* no client cert */
                                      NULL, /* No server cert callback */
                                      test_pool);
@@ -1560,8 +1658,8 @@ static void test_ssl_large_response(CuTest *tc)
                                      message_list, num_requests,
                                      action_list, num_requests, 0,
                                      https_set_root_ca_conn_setup,
-                                     "test/server/serfserverkey.pem",
-                                     server_certs,
+                                     get_srcdir_file(test_pool, "test/server/serfserverkey.pem"),
+                                     server_certs_srcdir(server_certs, test_pool),
                                      NULL, /* no client cert */
                                      NULL, /* No server cert callback */
                                      test_pool);
@@ -1599,8 +1697,8 @@ static void test_ssl_large_request(CuTest *tc)
                                      message_list, num_requests,
                                      action_list, num_requests, 0,
                                      https_set_root_ca_conn_setup,
-                                     "test/server/serfserverkey.pem",
-                                     server_certs,
+                                     get_srcdir_file(test_pool, "test/server/serfserverkey.pem"),
+                                     server_certs_srcdir(server_certs, test_pool),
                                      NULL, /* no client cert */
                                      NULL, /* No server cert callback */
                                      test_pool);
@@ -1623,7 +1721,7 @@ static apr_status_t client_cert_cb(void *data, const char **cert_path)
 
     tb->result_flags |= TEST_RESULT_CLIENT_CERTCB_CALLED;
 
-    *cert_path = "test/server/serfclientcert.p12";
+    *cert_path = get_srcdir_file(tb->pool, "test/server/serfclientcert.p12");
 
     return APR_SUCCESS;
 }
@@ -1635,8 +1733,8 @@ static apr_status_t client_cert_pw_cb(void *data,
     test_baton_t *tb = data;
 
     tb->result_flags |= TEST_RESULT_CLIENT_CERTPWCB_CALLED;
-    
-    if (strcmp(cert_path, "test/server/serfclientcert.p12") == 0)
+    if (strcmp(cert_path,
+               get_srcdir_file(tb->pool, "test/server/serfclientcert.p12")) == 0)
     {
         *password = "serftest";
         return APR_SUCCESS;
@@ -1695,8 +1793,8 @@ static void test_ssl_client_certificate(CuTest *tc)
                                      message_list, num_requests,
                                      action_list, num_requests, 0,
                                      client_cert_conn_setup,
-                                     "test/server/serfserverkey.pem",
-                                     all_server_certs,
+                                     get_srcdir_file(test_pool, "test/server/serfserverkey.pem"),
+                                     server_certs_srcdir(all_server_certs, test_pool),
                                      "Serf Client",
                                      NULL, /* No server cert callback */
                                      test_pool);
@@ -1718,7 +1816,6 @@ static void test_ssl_expired_server_cert(CuTest *tc)
     test_baton_t *tb;
     handler_baton_t handler_ctx[1];
     const int num_requests = sizeof(handler_ctx)/sizeof(handler_ctx[0]);
-    int expected_failures;
     apr_status_t status;
     test_server_message_t message_list[] = {
         {CHUNKED_REQUEST(1, "1")},
@@ -1740,21 +1837,22 @@ static void test_ssl_expired_server_cert(CuTest *tc)
                                      message_list, num_requests,
                                      action_list, num_requests, 0,
                                      NULL, /* default conn setup */
-                                     "test/server/serfserverkey.pem",
-                                     expired_server_certs,
+                                     get_srcdir_file(test_pool, "test/server/serfserverkey.pem"),
+                                     server_certs_srcdir(expired_server_certs, test_pool),
                                      NULL, /* no client cert */
-                                     ssl_server_cert_cb_expect_failures,
+                                     ssl_server_cert_cb_log,
                                      test_pool);
     CuAssertIntEquals(tc, APR_SUCCESS, status);
 
-    expected_failures = SERF_SSL_CERT_SELF_SIGNED |
-                        SERF_SSL_CERT_EXPIRED;
-    tb->user_baton = &expected_failures;
-
     create_new_request(tb, &handler_ctx[0], "GET", "/", 1);
 
     test_helper_run_requests_expect_ok(tc, tb, num_requests, handler_ctx,
                                        test_pool);
+    CuAssertStrEquals(tc,
+        "cert_cb: failures = CERT_SELF_SIGNED, cert = (CN=Serf Root CA, depth=2)\n"
+        "cert_cb: failures = CERT_EXPIRED, cert = (CN=localhost, depth=0)\n"
+        "cert_cb: failures = CERT_EXPIRED, cert = (CN=localhost, depth=0)\n",
+        tb->user_baton);
 }
 
 /* Validate that the expired certificate is reported as failure in the
@@ -1764,7 +1862,6 @@ static void test_ssl_future_server_cert(CuTest *tc)
     test_baton_t *tb;
     handler_baton_t handler_ctx[1];
     const int num_requests = sizeof(handler_ctx)/sizeof(handler_ctx[0]);
-    int expected_failures;
     apr_status_t status;
     test_server_message_t message_list[] = {
         {CHUNKED_REQUEST(1, "1")},
@@ -1786,21 +1883,22 @@ static void test_ssl_future_server_cert(CuTest *tc)
                                      message_list, num_requests,
                                      action_list, num_requests, 0,
                                      NULL, /* default conn setup */
-                                     "test/server/serfserverkey.pem",
-                                     future_server_certs,
+                                     get_srcdir_file(test_pool, "test/server/serfserverkey.pem"),
+                                     server_certs_srcdir(future_server_certs, test_pool),
                                      NULL, /* no client cert */
-                                     ssl_server_cert_cb_expect_failures,
+                                     ssl_server_cert_cb_log,
                                      test_pool);
     CuAssertIntEquals(tc, APR_SUCCESS, status);
 
-    expected_failures = SERF_SSL_CERT_SELF_SIGNED |
-                        SERF_SSL_CERT_NOTYETVALID;
-    tb->user_baton = &expected_failures;
-
     create_new_request(tb, &handler_ctx[0], "GET", "/", 1);
 
     test_helper_run_requests_expect_ok(tc, tb, num_requests, handler_ctx,
                                        test_pool);
+    CuAssertStrEquals(tc,
+        "cert_cb: failures = CERT_SELF_SIGNED, cert = (CN=Serf Root CA, depth=2)\n"
+        "cert_cb: failures = CERT_NOTYETVALID, cert = (CN=localhost, depth=0)\n"
+        "cert_cb: failures = CERT_NOTYETVALID, cert = (CN=localhost, depth=0)\n",
+        tb->user_baton);
 }
 
 
@@ -1853,8 +1951,8 @@ static void test_setup_ssltunnel(CuTest *tc)
                                            action_list_proxy, 2,
                                            0,
                                            https_set_root_ca_conn_setup,
-                                           "test/server/serfserverkey.pem",
-                                           server_certs,
+                                           get_srcdir_file(test_pool, "test/server/serfserverkey.pem"),
+                                           server_certs_srcdir(server_certs, test_pool),
                                            NULL, /* no client cert */
                                            NULL, /* No server cert callback */
                                            test_pool);
@@ -1897,7 +1995,6 @@ static void test_ssltunnel_no_creds_cb(CuTest *tc)
             "Transfer-Encoding: chunked" CRLF
             "Proxy-Authenticate: Basic realm=""Test Suite Proxy""" CRLF
             CRLF
-            "1" CRLF CRLF
             "0" CRLF CRLF},
     };
 
@@ -1914,8 +2011,8 @@ static void test_ssltunnel_no_creds_cb(CuTest *tc)
                                            action_list_proxy, 1,
                                            0,
                                            https_set_root_ca_conn_setup,
-                                           "test/server/serfserverkey.pem",
-                                           server_certs,
+                                           get_srcdir_file(test_pool, "test/server/serfserverkey.pem"),
+                                           server_certs_srcdir(server_certs, test_pool),
                                            NULL, /* no client cert */
                                            NULL, /* No server cert callback */
                                            test_pool);
@@ -1972,7 +2069,8 @@ ssltunnel_basic_authn_callback(char **username,
    tunnel. Retry the authentication a few times to test requeueing of the 
    CONNECT request. */
 static void ssltunnel_basic_auth(CuTest *tc, const char *server_resp_hdrs,
-                                 const char *proxy_resp_hdrs)
+                                 const char *proxy_407_resp_hdrs,
+                                 const char *proxy_200_resp_hdrs)
 {
     test_baton_t *tb;
     handler_baton_t handler_ctx[1];
@@ -2012,8 +2110,7 @@ static void ssltunnel_basic_auth(CuTest *tc, const char *server_resp_hdrs,
         "Proxy-Authenticate: Basic realm=""Test Suite Proxy""" CRLF
         "%s"
         CRLF
-        "1" CRLF CRLF
-        "0" CRLF CRLF, proxy_resp_hdrs);
+        "0" CRLF CRLF, proxy_407_resp_hdrs);
     action_list_proxy[1].kind = SERVER_RESPOND;
     action_list_proxy[1].text = apr_psprintf(test_pool,
         "HTTP/1.1 407 Unauthorized" CRLF
@@ -2021,8 +2118,7 @@ static void ssltunnel_basic_auth(CuTest *tc, const char *server_resp_hdrs,
         "Proxy-Authenticate: Basic realm=""Test Suite Proxy""" CRLF
         "%s"
         CRLF
-        "1" CRLF CRLF
-        "0" CRLF CRLF, proxy_resp_hdrs);
+        "0" CRLF CRLF, proxy_407_resp_hdrs);
 
     action_list_proxy[2].kind = SERVER_RESPOND;
     action_list_proxy[2].text = apr_psprintf(test_pool,
@@ -2031,11 +2127,13 @@ static void ssltunnel_basic_auth(CuTest *tc, const char *server_resp_hdrs,
         "Proxy-Authenticate: Basic realm=""Test Suite Proxy""" CRLF
         "%s"
         CRLF
-        "1" CRLF CRLF
-        "0" CRLF CRLF, proxy_resp_hdrs);
+        "0" CRLF CRLF, proxy_407_resp_hdrs);
 
     action_list_proxy[3].kind = SERVER_RESPOND;
-    action_list_proxy[3].text = CHUNKED_EMPTY_RESPONSE;
+    action_list_proxy[3].text = apr_psprintf(test_pool,
+        "HTTP/1.1 200 Connection Established" CRLF
+        "%s"
+        CRLF, proxy_200_resp_hdrs);
     /* Forward the remainder of the data to the server without validation */
     action_list_proxy[4].kind = PROXY_FORWARD;
     action_list_proxy[4].text = "https://localhost:" SERV_PORT_STR;
@@ -2074,7 +2172,6 @@ static void ssltunnel_basic_auth(CuTest *tc, const char *server_resp_hdrs,
         "WWW-Authenticate: Basic realm=""Test Suite""" CRLF
         "%s"
         CRLF
-        "1" CRLF CRLF
         "0" CRLF CRLF, server_resp_hdrs);
     action_list_server[1].kind = SERVER_RESPOND;
     action_list_server[1].text = CHUNKED_EMPTY_RESPONSE;
@@ -2090,8 +2187,8 @@ static void ssltunnel_basic_auth(CuTest *tc, const char *server_resp_hdrs,
                                            action_list_proxy, 7,
                                            0,
                                            https_set_root_ca_conn_setup,
-                                           "test/server/serfserverkey.pem",
-                                           server_certs,
+                                           get_srcdir_file(test_pool, "test/server/serfserverkey.pem"),
+                                           server_certs_srcdir(server_certs, test_pool),
                                            NULL, /* no client cert */
                                            NULL, /* No server cert callback */
                                            test_pool);
@@ -2121,19 +2218,25 @@ static void ssltunnel_basic_auth(CuTest *tc, const char *server_resp_hdrs,
 static void test_ssltunnel_basic_auth(CuTest *tc)
 {
     /* KeepAlive On for both proxy and server */
-    ssltunnel_basic_auth(tc, "", "");
+    ssltunnel_basic_auth(tc, "", "", "");
 }
 
 static void test_ssltunnel_basic_auth_server_has_keepalive_off(CuTest *tc)
 {
     /* Add Connection:Close header to server response */
-    ssltunnel_basic_auth(tc, "Connection: close" CRLF, "");
+    ssltunnel_basic_auth(tc, "Connection: close" CRLF, "", "");
 }
 
 static void test_ssltunnel_basic_auth_proxy_has_keepalive_off(CuTest *tc)
 {
-    /* Add Connection:Close header to proxy response */
-    ssltunnel_basic_auth(tc, "", "Connection: close" CRLF);
+    /* Add Connection:Close header to proxy 407 response */
+    ssltunnel_basic_auth(tc, "", "Connection: close" CRLF, "");
+}
+
+static void test_ssltunnel_basic_auth_proxy_close_conn_on_200resp(CuTest *tc)
+{
+    /* Add Connection:Close header to proxy 200 Conn. Establ. response  */
+    ssltunnel_basic_auth(tc, "", "", "Connection: close" CRLF);
 }
 
 static apr_status_t
@@ -2213,7 +2316,6 @@ static void test_ssltunnel_digest_auth(CuTest *tc)
             "nonce=\"ABCDEF1234567890\",opaque=\"myopaque\","
             "algorithm=\"MD5\",qop-options=\"auth\"" CRLF
             CRLF
-            "1" CRLF CRLF
             "0" CRLF CRLF},
         {SERVER_RESPOND, CHUNKED_EMPTY_RESPONSE},
         /* Forward the remainder of the data to the server without validation */
@@ -2233,8 +2335,8 @@ static void test_ssltunnel_digest_auth(CuTest *tc)
                                            action_list_proxy, 3,
                                            0,
                                            https_set_root_ca_conn_setup,
-                                           "test/server/serfserverkey.pem",
-                                           server_certs,
+                                           get_srcdir_file(test_pool, "test/server/serfserverkey.pem"),
+                                           server_certs_srcdir(server_certs, test_pool),
                                            NULL, /* no client cert */
                                            NULL, /* No server cert callback */
                                            test_pool);
@@ -2251,6 +2353,51 @@ static void test_ssltunnel_digest_auth(CuTest *tc)
     CuAssertTrue(tc, tb->result_flags & TEST_RESULT_AUTHNCB_CALLED);
 }
 
+/* Implements test_request_setup_t */
+static apr_status_t setup_request_err(serf_request_t *request,
+                                      void *setup_baton,
+                                      serf_bucket_t **req_bkt,
+                                      apr_pool_t *pool)
+{
+    static mockbkt_action actions[] = {
+        { 1, "a", APR_SUCCESS },
+        { 1, "", APR_EINVAL }
+    };
+    handler_baton_t *ctx = setup_baton;
+    serf_bucket_alloc_t *alloc;
+    serf_bucket_t *mock_bkt;
+
+    alloc = serf_request_get_alloc(request);
+    mock_bkt = serf_bucket_mock_create(actions, 2, alloc);
+    *req_bkt = serf_request_bucket_request_create(request,
+                                                  ctx->method, ctx->path,
+                                                  mock_bkt, alloc);
+    return APR_SUCCESS;
+}
+
+static void test_outgoing_request_err(CuTest *tc)
+{
+    test_baton_t *tb;
+    handler_baton_t handler_ctx[1];
+    apr_status_t status;
+    apr_pool_t *test_pool = tc->testBaton;
+
+    status = test_http_server_setup(&tb, NULL, 0,
+                                    NULL, 0, 0, NULL,
+                                    test_pool);
+    CuAssertIntEquals(tc, APR_SUCCESS, status);
+
+    create_new_request_ex(tb, &handler_ctx[0], "POST", "/", 1,
+                          setup_request_err, NULL);
+
+    status = test_helper_run_requests_no_check(tc, tb, 1, handler_ctx,
+                                               test_pool);
+    CuAssertIntEquals(tc, APR_EINVAL, status);
+    CuAssertIntEquals(tc, 1, tb->sent_requests->nelts);
+    CuAssertIntEquals(tc, 0, tb->accepted_requests->nelts);
+    CuAssertIntEquals(tc, 0, tb->handled_requests->nelts);
+}
+
 /*****************************************************************************/
 CuSuite *test_context(void)
 {
@@ -2286,7 +2433,9 @@ CuSuite *test_context(void)
     SUITE_ADD_TEST(suite, test_ssltunnel_basic_auth);
     SUITE_ADD_TEST(suite, test_ssltunnel_basic_auth_server_has_keepalive_off);
     SUITE_ADD_TEST(suite, test_ssltunnel_basic_auth_proxy_has_keepalive_off);
+    SUITE_ADD_TEST(suite, test_ssltunnel_basic_auth_proxy_close_conn_on_200resp);
     SUITE_ADD_TEST(suite, test_ssltunnel_digest_auth);
+    SUITE_ADD_TEST(suite, test_outgoing_request_err);
 
     return suite;
 }
